"""
Provides middleware to support classic CUL/Apache log format.

For accurate request metrics, apply this middleware before any others.

You should use the following log format when running uWSGI with this
middleware installed:

.. code-block::

   %(addr) %(addr) - %(user_id)|%(session_id) [%(rtime)] [%(uagent)] "%(method) %(uri) %(proto)" %(status) %(size) %(micros) %(ttfb) %(requestid)

"""

from typing import Type, Callable, List, Iterable, Tuple
from datetime import datetime
from pytz import timezone
from .base import BaseMiddleware

EASTERN = timezone('US/Eastern')

# The uwsgi package is injected into the Python environment by uWSGI, and is
# not available otherwise. If we can't import the package, there's not much
# else to do.
try:
    import uwsgi

    class ClassicLogsMiddleware(BaseMiddleware):
        """Add params to uwsgi to support classic CUL/Apache log format."""

        def before(self, environ: dict, start_response: Callable) \
                -> Tuple[dict, Callable]:
            """
            Inject request time in CUL classic format, and start TTFB clock.

            Parameters
            ----------
            environ : dict
                WSGI request environment.
            start_response : function
                Function used to begin the HTTP response. See
                https://www.python.org/dev/peps/pep-0333/#the-start-response-callable

            Returns
            -------
            dict
                Request environment (``environ``).
            function
                ``start_response`` function.
            """
            self.start = datetime.now()

            # Add the request ID as a uWSGI log variable.
            uwsgi.set_logvar('requestid', environ.get('REQUEST_ID', ''))

            # Express the time that the request was received in the classic
            # format.
            rtime = datetime.now(tz=EASTERN).strftime('%d/%b/%Y:%H:%M:%S %z')
            uwsgi.set_logvar('rtime', rtime)
            return environ, start_response

        def after(self, response: Iterable) -> Iterable:
            """
            Set TTFB log variable.

            Parameters
            ----------
            response : iterable
                Iterable that generates the HTTP response. See
                https://www.python.org/dev/peps/pep-0333/#the-application-framework-side

            Returns
            -------
            iterable
                Iterable that generates the HTTP response. See
                https://www.python.org/dev/peps/pep-0333/#the-application-framework-side
            """
            # This is a close approximation of "TTFB" (time from the
            # request received to the start of the response).
            ttfb = (datetime.now() - self.start).microseconds
            uwsgi.set_logvar('ttfb', str(ttfb).encode('ascii'))
            return response

except ImportError as ex:     # Not running in uWSGI.
    # BaseMiddleware does nothing.
    ClassicLogsMiddleware = BaseMiddleware  # type: ignore



"""Tests for arXiv taxonomy module."""
from unittest import TestCase
from datetime import date
from .definitions import GROUPS, ARCHIVES, \
    ARCHIVES_ACTIVE, CATEGORIES, ARCHIVES_SUBSUMED, \
    LEGACY_ARCHIVE_AS_PRIMARY, LEGACY_ARCHIVE_AS_SECONDARY, CATEGORY_ALIASES
from .category import Category, Archive, Group


class TestTaxonomy(TestCase):
    """Tests for the arXiv category taxonomy definitions."""

    def test_groups(self):
        """Tests for the highest level of the category taxonomy (groups)."""
        for key, value in GROUPS.items():
            self.assertRegexpMatches(key, r'^grp_[a-z\-_]+$')
            self.assertIn('name', value, 'name defined for {}'.format(key))
            self.assertIsInstance(value['name'], str, 'name is a str')
            self.assertIn(
                'start_year', value, 'start_year defined for {}'.format(key))
            self.assertIsInstance(
                value['start_year'], int, 'start_year is an integer')
            self.assertGreater(value['start_year'],
                               1990, 'start_year > 1990')
            if 'default_archive' in value:
                self.assertIn(
                    value['default_archive'],
                    ARCHIVES,
                    'default_archive {} is a valid archive'.format(
                        value['default_archive'])
                )

            group = Group(key)
            self.assertIsInstance(group.name, str)
            self.assertIsInstance(group.canonical, Category)
            self.assertIsInstance(group.display, str)
            self.assertIsInstance(group.unalias(), Category)

    def test_archives(self):
        """Tests for the middle level of the category taxonomy (archives)."""
        for key, value in ARCHIVES.items():
            self.assertIn('name', value, 'name defined for {}'.format(key))
            self.assertIsInstance(value['name'], str, 'name is a str')
            self.assertIn('in_group', value,
                          'in_group defined for {}'.format(key))
            self.assertIn(value['in_group'], GROUPS,
                          '{} is a valid group'.format(value['in_group']))
            self.assertIn('start_date', value,
                          'start_date defined for {}'.format(key))
            # start_dt = datetime.strptime(value['start_date'], '%Y-%m')
            self.assertIsInstance(value['start_date'], date)
            self.assertGreaterEqual(value['start_date'], date(1991, 8, 1))
            if 'end_date' in value:
                # end_dt = datetime.strptime(value['end_date'], '%Y-%m')
                self.assertIsInstance(value['end_date'], date)
                self.assertGreater(
                    value['end_date'],
                    value['start_date'],
                    'end_date greater than start_date'
                )

            archive = Archive(key)
            self.assertIsInstance(archive.name, str)
            self.assertIsInstance(archive.canonical, Category)
            self.assertIsInstance(archive.display, str)
            self.assertIsInstance(archive.unalias(), Category)

    def test_active_archives(self):
        """Tests for active (non-defunct) archives."""
        for key, value in ARCHIVES_ACTIVE.items():
            self.assertNotIn('end_date', value)

            archive = Archive(key)
            self.assertIsInstance(archive.name, str)
            self.assertIsInstance(archive.canonical, Category)
            self.assertIsInstance(archive.display, str)
            self.assertIsInstance(archive.unalias(), Category)

    def test_archives_subsumed(self):
        """Tests for defunct archives that have been subsumed by categories."""
        for key, value in ARCHIVES_SUBSUMED.items():
            self.assertIn(key, ARCHIVES, '{} is a valid archive'.format(key))
            self.assertIn(
                'end_date',
                ARCHIVES[key],
                '{} is a defunct archive'.format(key)
            )
            self.assertIn(
                value,
                CATEGORIES,
                '{} is a valid category'.format(value)
            )
            self.assertNotIn(
                'end_date',
                ARCHIVES[CATEGORIES[value]['in_archive']],
                '{} is not in a defunct archive'.format(value)
            )
            archive = Archive(key)
            self.assertIsInstance(archive.name, str)
            self.assertIsInstance(archive.canonical, Category)
            self.assertIsInstance(archive.display, str)
            self.assertIsInstance(archive.unalias(), Category)

    def test_legacy_archives_as_categories(self):
        """Test for archives that were used as primary/secondary categories."""
        for key, value in LEGACY_ARCHIVE_AS_PRIMARY.items():
            self.assertIn(key, ARCHIVES, '{} is a valid archive'.format(key))
            # dt = datetime.strptime(value, '%Y-%m')
            self.assertIsInstance(value, date)

            archive = Archive(key)
            self.assertIsInstance(archive.name, str)
            self.assertIsInstance(archive.canonical, Category)
            self.assertIsInstance(archive.display, str)
            self.assertIsInstance(archive.unalias(), Category)

        for key, value in LEGACY_ARCHIVE_AS_SECONDARY.items():
            self.assertIn(key, ARCHIVES, '{} is a valid archive'.format(key))
            # dt = datetime.strptime(value, '%Y-%m')
            self.assertIsInstance(value, date)

            archive = Archive(key)
            self.assertIsInstance(archive.name, str)
            self.assertIsInstance(archive.canonical, Category)
            self.assertIsInstance(archive.display, str)
            self.assertIsInstance(archive.unalias(), Category)

    def test_categories(self):
        """Test for the lowest level of the category taxonomy (categories)."""
        for key, value in CATEGORIES.items():
            self.assertIn('name', value, 'name defined for {}'.format(key))
            self.assertIsInstance(value['name'], str, 'name is a str')
            self.assertIn('in_archive', value,
                          'in_archive defined for {}'.format(key))
            self.assertIn(value['in_archive'], ARCHIVES,
                          '{} is a valid archive'.format(value['in_archive']))
            self.assertIn('is_active', value),
            self.assertIsInstance(value['is_active'], bool)

            category = Category(key)
            self.assertIsInstance(category.name, str)
            self.assertIsInstance(category.canonical, Category)
            self.assertIsInstance(category.display, str)
            self.assertIsInstance(category.unalias(), Category)

    def test_aliases(self):
        """Test for category aliases."""
        for key, value in CATEGORY_ALIASES.items():
            self.assertNotEqual(key, value,
                                'alias should be different from canonical')
            self.assertIn(key, CATEGORIES)
            self.assertIn(value, CATEGORIES)

            self.assertIn(f'{key} is an alias for {value}.',
                          CATEGORIES[key]['description'],
                          'Category description for an alias should indicate '
                          'that it is an alias for some other category')

            category = Category(key)
            self.assertIsInstance(category.unalias(), Category)
            self.assertNotEqual(category.unalias(), category)



"""
Provides a base class for HTTP API integrations.

The goal of this module is to handle boilerplate connection and request
handling that is invariant (or nearly so) among service integration modules
that target HTTP APIs.

.. todo::
   Make retry parameters easier to configure.


Specific goals
--------------

- Handle verifying the response status code, and raise appropriate and
  semantically meaningful exceptions.
- Handle setting up a persistent HTTP session to cut down on overhead.
- Provide basic retry functionality.
- Handle binding a service instance (with its persistent session) to the
  Flask application context.

"""

from typing import Optional, Tuple, MutableMapping, List, Any
from http import HTTPStatus as status
import inspect
import requests
from urllib3.util.retry import Retry
from urllib.parse import urlparse, urlunparse, urlencode
import json

from flask import g, Flask, current_app


from ..meta import MetaIntegration
import logging

from .exceptions import RequestFailed, RequestUnauthorized, RequestForbidden, \
    NotFound, BadRequest, SecurityException, ConnectionFailed, BadResponse

logger = logging.getLogger(__name__)

DEFAULT_ENDPOINT = 'http://fooservice:8000/'
DEFAULT_VERIFY = True


class HTTPIntegration(metaclass=MetaIntegration):
    """
    Integrate with a service over HTTP.

    Provides binding to the application context, so that connections can be
    re-used.

    To implement an HTTP API service integration, define a class that inherits
    from :class:`.HTTPIntegration`. At minimum, it should include an inner
    class called ``Meta`` with the attribute ``service_name``; this is used to
    identify the configuration parameters for the service. For example:

    .. code-block:: python

       class MyCoolIntegration(HTTPIntegration):
           class Meta:
               service_name = "cool_beans"

           def get_something(self, anything: str, token: str) -> dict:
               data, _, _ = self.json("get", f"/thing/{anything}", token)
               return data


    and in your ``config.py``, include:

    .. code-block:: python

       COOL_BEANS_ENDPOINT = "http://thecool.beans.arxiv.org:9876"
       COOL_BEANS_VERIFY = False   # (or True, if you are using SSL)


    Thanks to :class:`arxiv.integration.meta.MetaIntegration`, you should be
    able to use this integration in your app like:

    .. code-block:: python

       from my.cool.services import MyCoolIntegration
       from arxiv.users import auth

       app = Flask(__name__)
       MyCoolIntegration.init_app(app)
       auth.Auth(app)

       @app.route('/foo')
       def foo():
           mci = MyCoolIntegration.current_session()
           mci.get_something('foo', request.environ['token'])
           ...


    Any additional parameters in your app config that start with the
    ``service_name`` in upper case will be passed in as kwargs to the
    constructor. By default, these are stored on ``._extra`` for internal
    use.

    """

    class Meta:
        """
        Additional info about this class.

        This should be overridden by any child class.
        """

        service_name = "base"

    default_retry_config = Retry(
        total=10,
        read=10,
        connect=10,
        status=10,
        backoff_factor=0.5
    )
    """Default retry behavior for HTTP request."""

    def __init__(self, endpoint: str, verify: bool = True,
                 headers: dict = {}, **extra: Any) -> None:
        """
        Initialize an HTTP session.

        Parameters
        ----------
        endpoint : str
            Base service URI.
        verify : bool
            Whether or not SSL certificate verification should enforced.
        headers : dict
            Headers to be included on all requests.
        extra : kwargs

        """
        self._extra = extra
        self._session = requests.Session()
        self._verify = verify
        self._retry = self.get_retry_config()
        self._adapter = requests.adapters.HTTPAdapter(max_retries=self._retry)
        self._session.mount(f'{urlparse(endpoint).scheme}://', self._adapter)
        if not endpoint.endswith('/'):
            endpoint += '/'
        self._endpoint = endpoint
        self._session.headers.update(headers)

    def get_retry_config(self) -> Retry:
        """
        Defines the retry behavior for HTTP requests.

        This is defined as an instance method so that it can be defined in
        relation to an app config or other contextual information.

        See :class:`urllib3.util.retry.Retry`.
        """
        return self.default_retry_config

    def _path(self, path: str, query: dict = {}) -> str:
        """Generate a full path for a request."""
        o = urlparse(self._endpoint)
        path = path.lstrip('/')
        return urlunparse((  # type: ignore
            o.scheme, o.netloc, f"{o.path}{path}",
            None, urlencode(query), None
        ))

    def _check_status(self, resp: requests.Response,
                      expected_code: List[int] = [status.OK]) -> None:
        """Check for unexpected or errant status codes."""
        if resp.status_code not in expected_code:
            raise_for_http_status(resp.status_code, resp)
            raise RequestFailed(f'Unexpected code: {resp.status_code}', resp)

    def _parse_location(self, location: str) -> str:
        endpoint_path = urlparse(self._endpoint).path
        if self._endpoint in location:
            _, location = location.split(self._endpoint, 1)
        elif endpoint_path in location:
            _, location = location.split(endpoint_path, 1)
        return location

    def request(self, method: str, path: str, token: Optional[str] = None,
                expected_code: List[int] = [status.OK],
                allow_2xx_redirects: bool = True,
                **kwargs: Any) -> requests.Response:
        """Make an HTTP request with error/code handling."""
        if token is not None:
            if 'headers' not in kwargs:
                kwargs['headers'] = {}
            kwargs['headers'].update({'Authorization': token})
        resp: requests.Response
        try:
            make_request = getattr(self._session, method)
            resp = make_request(self._path(path), **kwargs)
            logger.debug('Got response %s', resp)
        except requests.exceptions.SSLError as ex:
            raise SecurityException('SSL failed: %s' % ex) from ex
        except requests.exceptions.ConnectionError as ex:
            raise ConnectionFailed('Could not connect: %s' % ex) from ex

        # 200-series redirects are a common pattern on our projects, so these
        # should be supported.
        is_2xx = status.OK < resp.status_code \
            and resp.status_code < status.MULTIPLE_CHOICES
        if allow_2xx_redirects and is_2xx and 'Location' in resp.headers:
            loc = self._parse_location(resp.headers['Location'])
            logger.debug('Following 2xx redirect to %s', loc)
            resp = self.request('get', loc, token, expected_code,
                                allow_2xx_redirects)

        self._check_status(resp, expected_code)
        return resp

    def json(self, method: str, path: str, token: Optional[str] = None,
             expected_code: List[int] = [status.OK], **kwargs: Any) \
            -> Tuple[dict, int, MutableMapping]:
        """
        Perform an HTTP request to a JSON endpoint, and handle any exceptions.

        Returns
        -------
        dict
            Response content.
        int
            HTTP status code.
        dict
            Response headers.

        """
        response = self.request(method, path, token, expected_code, **kwargs)
        try:
            return response.json(), response.status_code, response.headers
        except json.decoder.JSONDecodeError as ex:
            raise BadResponse(f'Could not decode', response) from ex

    def get_status(self) -> dict:
        """Get the status of the service."""
        logger.debug('Get service status')
        content = self.json('get', 'status')[0]
        logger.debug('Got status response: %s', content)
        return content

    @classmethod
    def init_app(cls, app: Flask) -> None:
        """Set default configuration parameters for an application instance."""
        if not hasattr(cls, 'Meta'):
            raise NotImplementedError('Child class must have Meta class')
        name = cls.Meta.service_name.upper()
        app.config.setdefault(f'{name}_ENDPOINT', DEFAULT_ENDPOINT)
        app.config.setdefault(f'{name}_VERIFY', DEFAULT_VERIFY)

    @classmethod
    def get_session(cls, app: Optional[Flask] = None) -> 'HTTPIntegration':
        """Get a new session with the service."""
        if not hasattr(cls, 'Meta'):
            raise NotImplementedError('Child class must have Meta class')
        if app is None:
            app = current_app
        name = cls.Meta.service_name.upper()
        try:
            params = app.config.get_namespace(f'{name}_')
            endpoint = params.pop('endpoint')
            verify = params.pop('verify')
        except KeyError as ex:
            raise RuntimeError('Must call init_app() on app before use') from ex

        logger.debug('Create %s session at endpoint %s', name, endpoint)
        return cls(endpoint, verify=verify, **params)

    @classmethod
    def current_session(cls) -> 'HTTPIntegration':
        """Get or create an HTTP session for this context."""
        name = cls.Meta.service_name
        if not g:
            return cls.get_session()
        elif name not in g:
            setattr(g, name, cls.get_session())
        return getattr(g, name)  # type: ignore


def raise_for_http_status(status_code: int,
                          resp: Optional[requests.Response] = None) -> None:
    """
    Raise an exception based on an HTTP status code.

    Parameters
    ----------
    status_code : int
        HTTP status code.
    resp : :class:`requests.Response`
        If provided, passed to the exception.

    Raises
    ------
    :class:`RequestFailed`
        Or one of its children. See :class:`.exceptions`.

    """
    if resp is None:
        raise ValueError('No response')
    if resp.status_code >= status.INTERNAL_SERVER_ERROR:
        raise RequestFailed(f'Status: {resp.status_code}', resp)
    elif resp.status_code == status.UNAUTHORIZED:
        raise RequestUnauthorized(f'Not authorized', resp)
    elif resp.status_code == status.FORBIDDEN:
        raise RequestForbidden(f'Forbidden', resp)
    elif resp.status_code == status.NOT_FOUND:
        raise NotFound(f'No such resource: {resp.url}', resp)
    elif resp.status_code >= status.BAD_REQUEST:
        raise BadRequest(f'Bad request: {resp.content}', resp)



#  Created by tag_check.write_version
#  NEVER CHECK THIS version.py file INTO GIT.
#  It should be generated when the package is build for distribution.
__when__ = '16/03/2020 18:49:34'
__version__ = '0.0.1-dev1'



"""Exceptions for HTTP API integrations."""

from typing import Optional
from requests import Response


class RequestFailed(IOError):
    """The file management service returned an unexpected status code."""

    def __init__(self, msg: str, response: Response) -> None:
        """Attach (optional) data to the exception."""
        self.response = response
        super(RequestFailed, self).__init__(msg)

    @property
    def status_code(self) -> Optional[int]:
        """Get the code from the response originating the exception."""
        if self.response:
            return self.response.status_code
        return None


class RequestUnauthorized(RequestFailed):
    """Client/user is not authenticated."""


class RequestForbidden(RequestFailed):
    """Client/user is not allowed to perform this request."""


class BadRequest(RequestFailed):
    """The request was malformed or otherwise improper."""


class Oversize(BadRequest):
    """The upload was too large."""


class NotFound(BadRequest):
    """The referenced upload does not exist."""


class BadResponse(RequestFailed):
    """The response from the file management service was malformed."""


class ConnectionFailed(IOError):
    """Could not connect to the file management service."""


class SecurityException(ConnectionFailed):
    """Raised when SSL connection fails."""



"""
arXiv base Flask components.

Provides :class:`.Base`, which attaches base templates, static assets, global
context processors, and exception handlers to a :class:`flask.Flask` app
instance.

Intended for use in an application factory. For example:

.. code-block:: python

   python
   from flask import Flask
   from arxiv.base import Base
   from someapp import routes


   def create_web_app() -> Flask:
      app = Flask('someapp')
      app.config.from_pyfile('config.py')

      Base(app)   # Registers the base/UI blueprint.
      app.register_blueprint(routes.blueprint)    # Your blueprint.
   return app


"""
import types
from typing import Optional, Any, Dict
from flask import Blueprint, Flask, Blueprint
from werkzeug.exceptions import NotFound

from . import exceptions, urls, alerts, context_processors, filters
from . import config as base_config
from .converter import ArXivConverter


class Base(object):
    """Attaches a base UI blueprint and context processors to an app."""

    def __init__(self, app: Optional[Flask] = None) -> None:
        """Initialize ``app`` with base blueprint."""
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Create and register the base UI blueprint."""
        # Attach the arXiv identifier converter for URLs with IDs.
        app.url_map.converters['arxiv'] = ArXivConverter

        # Set the static url path.
        app_version = app.config.get("APP_VERSION", "null")
        app_static_path = f'/static/{app.name}/{app_version}'
        app.static_url_path = app_static_path

        # We also need to update the app's url_map, since that's how url_for()
        # builds URLs.
        # Remove the rule from the url_map.
        for rule in app.url_map.iter_rules('static'):
            app.url_map._rules.remove(rule)
        # Werkzeug maintains this for faster lookup; we need to clear it, too.
        app.url_map._rules_by_endpoint['static'] = []
        # Add the updated rule.
        app.add_url_rule(f'{app.static_url_path}/<path:filename>',
                         endpoint='static',
                         view_func=app.send_static_file)

        base_static_url_path = f'/static/base/{base_config.BASE_VERSION}'

        # In some cases, we want an app to handle all of its static files, e.g.
        # when deploying in a development environment. Setting the
        # configuration param RELATIVE_STATIC_PATHS to True will cause
        # ``/{RELATIVE_STATIC_PATHS}`` to be prepended to the static paths for
        # base assets. This should have no impact on static paths for
        # blueprints.
        if app.config.get('RELATIVE_STATIC_PATHS'):
            prefix = app.config.get('RELATIVE_STATIC_PREFIX', '').strip('/')
            base_static_url_path = f'/{prefix}{base_static_url_path}'

        # The base blueprint attaches static assets and templates. These are
        # used by many different apps.
        #
        # We include the version of this package in the static path, so that
        # apps can run slightly different versions of this package without
        # clobbering each other.
        blueprint = Blueprint(
            'base',
            __name__,
            template_folder='templates',
            static_folder='static',
            static_url_path=base_static_url_path
        )
        app.register_blueprint(blueprint)

        # 2019-01-31 - Erick P.:
        #
        # We need to able able to set the static URL for blueprints, so that
        # we can ensure a consistent URL strategy. Changing the static URL for
        # a blueprint after it is registered won't work, because the blueprint
        # will have already set up its internal machinery via
        # Blueprint.register(), which is called by Flask.register_blueprint().
        # In order to intercept and modify the blueprint before
        # Blueprint.register() is called, we swap out the app's
        # register_blueprint() method for a closure (below) that alters the
        # static URL path.
        def register_blueprint(self: Flask, blueprint: Blueprint,
                               **options: Any) -> None:
            blueprint.static_url_path = f'{app_static_path}/{blueprint.name}'
            self._register_blueprint(blueprint, **options)

        app._register_blueprint = app.register_blueprint
        app.register_blueprint = types.MethodType(register_blueprint, app)    # type: ignore

        # Register base exception handlers.
        for error, handler in exceptions.get_handlers():
            app.errorhandler(error)(handler)

        # Attach the external URL handler as a fallback for failed calls to
        # url_for().
        urls.register_external_urls(app)

        filters.register_filters(app)
        context_processors.register_context_processors(app)



"""
Provides base classes and tools for building agents (Kinesis consumers).

We use Kinesis streams to broker notifications that concern multiple services
in the arXiv system. For example, at publication time, the publication process
generates notifications about new arXiv papers so that services like search
can update themselves.

This module provides a base class for building agents with Kinesis streams,
:class:`.BaseConsumer`. The objective is to provide all of the
stream-management (e.g. checkpointing) and error-handling boilerplate needed
for any Kinesis consumer, so that we can focus on building the idiosyncratic
functional bits in arXiv services.

Using :class:`.BaseConsumer`
============================

You (the developer) should be able to create a minimal agent by:

1. Defining a class that inherits from :class:`.BaseConsumer`
2. Defining an instace method on that class with the signature:
   ``def process_record(self, record: dict) -> None:`` that implements
   application-specific processing for each notification.
3. Calling :func:`.process_stream` with your consumer class and an application
   config (e.g. from Flask).

Your config should include the following:

``KINESIS_STREAM`` : str
    Name of the stream to consume.
``KINESIS_SHARD_ID`` : str
    E.g. ``"shardId-000000000000"``.
``AWS_ACCESS_KEY_ID`` : str
    Access key ID with read access to Kinesis streams.
``AWS_SECRET_ACCESS_KEY`` : str
    Secret access key with read access to Kinesis streams.
``AWS_REGION`` : str
    E.g. ``us-east-1``.
``KINESIS_ENDPOINT`` : str
    This should be ``None`` in production, but can be set to something else
    for integration testing (e.g. with localstack).
``KINESIS_VERIFY`` : str
    This should be ``"true"`` in production, but can be disabled when doing
    integration testing (since localstack certs won't verify).
``KINESIS_START_TYPE``: str
    How to get the first shard iterator (if a starting position is not
    available via the checkpointer). Currently supports ``TRIM_HORIZON``
    and ``AT_TIMESTAMP``.
``KINESIS_START_AT`` : str
    If using ``AT_TIMESTAMP``, the point of time from which to start in the
    stream. Should have the format ``'%Y-%m-%dT%H:%M:%S'``.

If you're using the provided :class:`.DiskCheckpointManager` provided here
(used by default in :func:`.process_stream`), you should also set:

``KINESIS_CHECKPOINT_VOLUME`` : str
    Full path to a directory where the consumer should store its position.
    Must be readable/writeable.


Testing and development
=======================

The easiest way to write tests for Kinesis consumers is to mock the
[``boto3.client`` factory](http://boto3.readthedocs.io/en/latest/reference/services/kinesis.html).
Unit tests for this module can be found in :mod:`arxiv.base.agent.tests`,
most of which mock boto3 in this way.

For integration tests, or developing against a "live" Kinesis stream,
[Localstack](https://github.com/localstack/localstack) provides a Kinesis
for testing/development purposes (port 4568). You can use the config
parameters above to point to a local instance of localstack (e.g. run with
Docker).

"""

import time
import json
from pytz import UTC
from datetime import datetime, timedelta
import os
from typing import Any, Optional, Tuple, Generator, Callable, Dict, Union, Any
from contextlib import contextmanager
import signal
import warnings

import boto3
from botocore.exceptions import WaiterError, NoCredentialsError, \
    PartialCredentialsError, BotoCoreError, ClientError

from retry.api import retry_call

import logging
from .exceptions import CheckpointError, StreamNotAvailable, StopProcessing, \
    KinesisRequestFailed, ConfigurationError, RestartProcessing

logger = logging.getLogger(__name__)
# logger.propagate = False
logger.setLevel(10)

NOW = datetime.now(tz=UTC).strftime('%Y-%m-%dT%H:%M:%S')


class DiskCheckpointManager(object):
    """
    Provides on-disk loading and updating of consumer checkpoints.

    You can substitute any other mechanism that you prefer for checkpointing
    when you instantiate your consumer, so long as the passed object:

    1. Has an instance method ``checkpoint(self, position: str) -> None:`` that
       stores ``position``, and
    2. Exposes a property ``.position`` that is the last value passed to
       ``checkpoint``.

    """

    def __init__(self, base_path: str, stream_name: str, shard_id: str) \
            -> None:
        """Load or create a new file for checkpointing."""
        if not os.path.exists(base_path):
            raise ValueError(f'Path does not exist: {base_path}')
        self.file_path = os.path.join(base_path,
                                      f'{stream_name}__{shard_id}.json')
        if not os.path.exists(self.file_path):
            try:
                with open(self.file_path, 'w') as f:
                    f.write('')
            except Exception as ex:   # The containing path doesn't exist.
                raise ValueError(f'Could not use {self.file_path}') from ex

        with open(self.file_path) as f:
            position = f.read()
        self.position = position if position else None

    def checkpoint(self, position: str) -> None:
        """Checkpoint at ``position``."""
        try:
            with open(self.file_path, 'w') as f:
                f.write(position)
            self.position = position
        except Exception as ex:
            raise CheckpointError('Could not checkpoint') from ex


class BaseConsumer(object):
    """
    Kinesis stream consumer.

    Consumes a single shard from a single stream, and checkpoints on disk
    (to reduce external dependencies).
    """

    def __init__(self, stream_name: str = '', shard_id: str = '',
                 access_key: Optional[str] = None,
                 secret_key: Optional[str] = None,
                 region: str = '',
                 checkpointer: Optional[DiskCheckpointManager] = None,
                 back_off: int = 5, batch_size: int = 50,
                 endpoint: Optional[str] = None, verify: bool = True,
                 duration: Optional[int] = None,
                 start_type: str = 'AT_TIMESTAMP',
                 start_at: str = NOW, tries: int = 5, delay: int = 5,
                 max_delay: Optional[int] = None, backoff: int = 1,
                 jitter: Union[int, Tuple[int, int]] = 0, **extra: Any) -> None:
        """Initialize a new stream consumer."""
        logger.info(f'New consumer for {stream_name} ({shard_id})')
        self.stream_name = stream_name
        self.shard_id = shard_id
        self.checkpointer = checkpointer
        if self.checkpointer:
            self.position = self.checkpointer.position
        else:
            self.position = None
        self.duration = duration
        self.start_time: Optional[float] = None
        self.back_off = back_off
        self.batch_size = batch_size
        self.sleep_time = 1
        self.start_at = start_at
        self.start_type = start_type
        self._access_key = access_key
        self._secret_key = secret_key
        logger.info(f'Got start_type={start_type} and start_at={start_at}')

        # Retry parameters.
        self.retry_params = {
            'tries': tries,
            'delay': delay,
            'max_delay': max_delay,
            'backoff': backoff,
            'jitter': jitter  #  extra seconds added to delay between retry attempts.
        }

        if not self.stream_name or not self.shard_id:
            logger.info('No stream set; no attempt to connect to Kinesis')
            raise RuntimeError('Stream and shard must be specified')

        self.endpoint = endpoint
        self.verify = verify
        self.region = region

        # Intercept SIGINT and SIGTERM so that we can checkpoint before exit.
        self.exit = False
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)
        logger.info('Ready to start')

    def get_or_create_stream(self) -> None:
        """Wait for the stream, and create it if it doesn't exist."""
        logger.info(f'Waiting for {self.stream_name} to be available')
        try:
            self.wait_for_stream(tries=1)
        except (KinesisRequestFailed, StreamNotAvailable):
            logger.info('Could not connect to stream; attempting to create')
            self.client.create_stream(StreamName=self.stream_name,
                                      ShardCount=1)
            logger.info(f'Created; waiting for {self.stream_name} again')
            self.wait_for_stream(**self.retry_params)   # type: ignore

    def stop(self, signal: int, frame: Any) -> None:
        """Set exit flag for a graceful stop."""
        logger.error(f'Received signal {signal}')
        self._checkpoint()
        logger.error('Done')
        raise StopProcessing(f'Received signal {signal}')

    def new_client(self) -> boto3.client:
        """Generate a new Kinesis client."""
        params: Dict[str, Any] = {'region_name': self.region,
                                  'aws_access_key_id': self._access_key,
                                  'aws_secret_access_key': self._secret_key}
        client_params: Dict[str, Any] = {}
        if self.endpoint:
            client_params['endpoint_url'] = self.endpoint
        if self.verify is False:
            client_params['verify'] = False

        logger.debug('New session with parameters: %s', params)
        # We don't want to let boto3 manage the Session for us.
        self._session = boto3.Session(**params)

        return self._session.client('kinesis', **client_params)

    def wait_for_stream(self, tries: int = 5, delay: int = 5,
                        max_delay: Optional[int] = None, backoff: int = 2,
                        jitter: Union[int, Tuple[int, int]] = 0) -> None:
        """
        Wait for the stream to become available.

        If the stream becomes available, returns ``None``. Otherwise, raises
        a :class:`.StreamNotAvailable` exception.

        Raises
        ------
        :class:`.StreamNotAvailable`
            Raised when the stream could not be reached.

        """
        waiter = self.client.get_waiter('stream_exists')
        try:
            logger.error(f'Waiting for stream {self.stream_name}')
            fkwargs = {
                'StreamName': self.stream_name,
                'WaiterConfig': {
                    'Delay': 1,
                    'MaxAttempts': 10
                }
            }
            retry_call(waiter.wait, fkwargs=fkwargs,
                       tries=tries, delay=delay, max_delay=max_delay,
                       backoff=backoff, jitter=jitter)
        except WaiterError as ex:
            logger.error('Failed to get stream while waiting')
            raise StreamNotAvailable('Could not connect to stream') from ex
        except (PartialCredentialsError, NoCredentialsError) as ex:
            logger.error('Credentials missing or incomplete: %s', ex)
            raise ConfigurationError('Credentials missing') from ex
        except ClientError as ex:
            code = ex.response['Error']['Code']
            raise KinesisRequestFailed(f'Last code: {code}') from ex

    def _get_iterator(self) -> str:
        """
        Get a new shard iterator.

        If our position is set, we will start with the record immediately after
        that position. Otherwise, we start at ``start_at`` timestamp.

        Returns
        -------
        str
            The sequence ID of the record on which to start.

        """
        params: Dict[str, Any] = dict(
            StreamName=self.stream_name,
            ShardId=self.shard_id
        )
        if self.position:
            params.update(dict(
                ShardIteratorType='AFTER_SEQUENCE_NUMBER',
                StartingSequenceNumber=self.position
            ))
        elif self.start_type == 'AT_TIMESTAMP' and self.start_at:
            start_at = datetime.strptime(self.start_at, '%Y-%m-%dT%H:%M:%S')
            params.update(dict(
                ShardIteratorType='AT_TIMESTAMP',
                Timestamp=(
                    start_at - datetime.utcfromtimestamp(0)
                ).total_seconds()
            ))
        elif self.start_type == 'TRIM_HORIZON':
            params.update(dict(ShardIteratorType='TRIM_HORIZON'))
        try:
            it: str = self.client.get_shard_iterator(**params)['ShardIterator']
            return it
        except self.client.exceptions.InvalidArgumentException as ex:
            logger.info('Got InvalidArgumentException: %s', str(ex))
            # Iterator may not have come from this stream/shard.
            if self.position is not None:
                self.position = None
                return self._get_iterator()
        raise KinesisRequestFailed('Could not get shard iterator')

    def _checkpoint(self) -> None:
        """
        Checkpoint at the current position.

        The current position is the sequence number of the last record that was
        successfully processed.
        """
        if self.position is not None and self.checkpointer:
            self.checkpointer.checkpoint(self.position)
            logger.debug(f'Set checkpoint at {self.position}')

    def get_records(self, iterator: str, limit: int, tries: int = 5,
                    delay: int = 5, max_delay: Optional[int] = None,
                    backoff: int = 1, jitter: Union[int, Tuple[int, int]] = 0)\
            -> Tuple[str, dict]:
        """Get the next batch of ``limit`` or fewer records."""
        logger.debug(f'Get more records from {iterator}, limit {limit}')
        fkwargs = dict(ShardIterator=iterator, Limit=limit)
        try:
            response = retry_call(self.client.get_records, fkwargs=fkwargs,
                                  exceptions=ClientError, tries=tries,
                                  delay=delay, max_delay=max_delay,
                                  backoff=backoff, jitter=jitter)
        except ClientError as ex:
            code = ex.response['Error']['Code']
            raise KinesisRequestFailed(f'Last code: {code}') from ex
        iterator = response['NextShardIterator']
        return iterator, response

    def _check_timeout(self) -> None:
        """If a processing duration is set, exit if duration is exceeded."""
        if not self.start_time or not self.duration:
            return
        running_for = time.time() - self.start_time
        if running_for > self.duration:
            logger.info(f'Ran for {running_for} seconds; exiting')
            self._checkpoint()
            raise StopProcessing(f'Ran for {running_for} seconds; exiting')

    def process_records(self, start: str) -> Tuple[str, int]:
        """Retrieve and process records starting at ``start``."""
        logger.debug(f'Get more records, starting at {start}')
        processed = 0
        try:
            time.sleep(self.sleep_time)   # Don't get carried away.
            next_start, response = self.get_records(start, self.batch_size,   # type: ignore
                                                    **self.retry_params)
        except Exception as ex:
            self._checkpoint()
            raise StopProcessing('Unhandled exception: %s' % str(ex)) from ex

        logger.debug('Got %i records', len(response['Records']))
        for record in response['Records']:
            self._check_timeout()

            # It is possible that Kinesis will replay the same message several
            # times, especially at the end of the stream. There's no point in
            # replaying the message, so we'll continue on.
            if record['SequenceNumber'] == self.position:
                continue

            self.process_record(record)
            processed += 1

            # Setting the position means that we have successfully
            # processed this record.
            if record['SequenceNumber']:    # Make sure it's set.
                self.position = record['SequenceNumber']
                logger.debug(f'Updated position to {self.position}')
        logger.debug(f'Next start is {next_start}')
        return next_start, processed

    def go(self) -> None:
        """Run the main processing routine."""
        self.client = self.new_client()
        self.get_or_create_stream()

        self.start_time = time.time()
        logger.info(f'Starting processing from position {self.position}'
                    f' on stream {self.stream_name} and shard {self.shard_id}')

        start = self._get_iterator()
        while True:
            start, processed = self.process_records(start)
            if processed > 0:
                self._checkpoint()  # Checkpoint after every batch.
            if start is None:     # Shard is closed.
                logger.error('Shard closed unexpectedly; no new iterator')
                self._checkpoint()
                raise StopProcessing('Could not get a new iterator')
            self._check_timeout()

    def process_record(self, record: dict) -> None:
        """
        Process a single record from the stream.

        Parameters
        ----------
        record : dict

        """
        logger.info(f'Processing record {record["SequenceNumber"]}')
        logger.debug(f'Process record {record}')
        # raise NotImplementedError('Should be implemented by a subclass')


def process_stream(Consumer: type, config: dict,
                   checkpointmanager: Optional[Any] = None,
                   duration: Optional[int] = None,
                   extra: Dict[str, Any] = {}) -> None:
    """
    Configure and run an agent (Kinesis consumer).

    Parameters
    ----------
    Consumer : type
        A class that inherits from :class:`.BaseConsumer`.
    config : dict
        An application config (e.g. a Flask config).
    duration : int
        Time (in seconds) to consume records. If None (default), will
        run "forever".
    extra : kwargs
        Extra keyword arguments passed to the Consumer constructor.

    """
    # By default we use the on-disk checkpoint manager.
    if checkpointmanager is None:
        checkpointmanager = DiskCheckpointManager(
            config['KINESIS_CHECKPOINT_VOLUME'],
            config['KINESIS_STREAM'],
            config['KINESIS_SHARD_ID'],
        )

    with warnings.catch_warnings():     # boto3 is notoriously annoying.
        warnings.simplefilter("ignore")
        start_at = config.get('KINESIS_START_AT')
        start_type = config.get('KINESIS_START_TYPE')
        if not start_type:
            start_type = 'AT_TIMESTAMP'
        if start_type == 'AT_TIMESTAMP' and not start_at:
            start_at = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')

        processor = Consumer(
            config['KINESIS_STREAM'],
            config['KINESIS_SHARD_ID'],
            config['AWS_ACCESS_KEY_ID'],
            config['AWS_SECRET_ACCESS_KEY'],
            config['AWS_REGION'],
            checkpointmanager,
            endpoint=config.get('KINESIS_ENDPOINT', None),
            verify=config.get('KINESIS_VERIFY', 'true') == 'true',
            duration=duration,
            start_type=start_type,
            start_at=start_at,
            tries=config.get('KINESIS_RETRY_TRIES', 5),
            delay=config.get('KINESIS_RETRY_DELAY', 5),
            max_delay=config.get('KINESIS_RETRY_MAX_DELAY', None),
            backoff=config.get('KINESIS_RETRY_BACKOFF', 1),
            jitter=config.get('KINESIS_RETRY_JITTER', 0),
            **extra
        )
        try:
            retry_call(processor.go, exceptions=RestartProcessing)
        except StopProcessing:
            logger.info('Got StopProcessing; stopping.')
            return



"""Provides base classes for WSGI middlewares."""

from typing import Callable, Union, Iterable, TypeVar, Optional, Mapping, Tuple
from typing_extensions import Protocol


WSGIRequest = Tuple[dict, Callable]
"""
The WSGI request.

Comprised of an environ mapping and a callable for starting the response. See
also `https://www.python.org/dev/peps/pep-0333/#the-start-response-callable`_.
"""

WSGIResponse = Iterable
"""
The iterable that generates a WSGI response.

See also
`https://www.python.org/dev/peps/pep-0333/#the-application-framework-side`_.
"""

IWSGIApp = Callable[[dict, Callable], WSGIResponse]


class IWSGIMiddleware(Protocol):
    """Defines a minimal class that can be used as a middleware."""

    def __init__(self, wsgi_app: IWSGIApp, config: Mapping = {}) -> None:
        """Initialize with an WSGI app and an optional configuration."""
        ...

    def __call__(self, environ: dict, start: Callable) -> WSGIResponse:
        """Support the WSGI protocol."""
        ...

    @property
    def wsgi_app(self) -> IWSGIApp:
        """Offer a ``wsgi_app`` property, per :class:`.Flask` behavior."""
        ...


class IWSGIMiddlewareFactory(Protocol):
    """Defines a minimal WSGI middleware factory."""

    def __call__(self, app: IWSGIApp, config: Mapping = {}) -> IWSGIMiddleware:
        """Generate a :class:`.WSGIMiddleware`."""
        ...


class BaseMiddleware:
    r"""
    Base class for WSGI middlewares.

    Child classes should override :func:`.before` and/or :func:`.after`\.
    """

    def __init__(self, wsgi_app: IWSGIApp, config: Mapping = {}) -> None:
        """
        Set the app factory that this middleware wraps.

        Parameters
        ----------
        wsgi_app : callable
            The application wrapped by this middleware. This might be an inner
            middleware, or the original :class:`.Flask` app itself.
        config : dict
            Application configuration.

        """
        self.app = wsgi_app
        self.config = config

    def before(self, environ: dict, start_response: Callable) -> WSGIRequest:
        """
        Pre-process a WSGI request. To be overridden by a child class.

        Parameters
        ----------
        environ : dict
            WSGI request environ.
        start : callable
            Callable used to begin the HTTP response.

        Returns
        -------
        dict
            WSGI request environ.
        callable
            Callable used to begin the HTTP response.

        """
        return environ, start_response

    def after(self, response: WSGIResponse) -> WSGIResponse:
        """
        Post-process a WSGI response. To be overridden by a child class.

        Parameters
        ----------
        response : iterable
            The WSGI response.

        Returns
        -------
        iterable
            The WSGI response.

        """
        return response

    def __call__(self, environ: dict, start: Callable) -> WSGIResponse:
        r"""
        Handle a WSGI request.

        Parameters
        ----------
        environ : dict
            WSGI request environment.
        start : function
            Function used to begin the HTTP response. See
            :const:`.WSGIRequest`\.

        Returns
        -------
        iterable
            Iterable that generates the HTTP response. See
            :const:`.WSGIResponse`\.

        """
        environ, start_response = self.before(environ, start)
        response: WSGIResponse = self.app(environ, start)
        response = self.after(response)
        return response

    @property
    def wsgi_app(self) -> IWSGIApp:
        """
        Refer to the current instance of this class.

        This is here for consistency with the :class:`.Flask` interface.
        """
        return self



"""
Provides base HTTP exceptions and exception handling.

arXiv-base exceptions inherit from `Werkzeug HTTPExceptions
<http://werkzeug.pocoo.org/docs/0.14/exceptions/>`_. In most cases it will
make sense to simply use Werkzeug's many HTTP exceptions directly. Custom
exceptions can also be defined by further subclassing those exceptions.

These handlers are attached to any Flask app passed to
:class:`arxiv.base.Base`. Thus any services that extends arXiv base need only
define (if needed) and raise exceptions.

Controllers should raise HTTP exceptions. The exception message (description)
will be rendered within the base template for that exception class, including
markup.
"""

from typing import Callable, List, Tuple, Type, Dict

from werkzeug.exceptions import NotFound, Forbidden, Unauthorized, \
    MethodNotAllowed, RequestEntityTooLarge, BadRequest, InternalServerError, \
    HTTPException, default_exceptions
from flask import render_template, make_response, Response, request

from http import HTTPStatus as status

_handlers: List[Tuple[Type[HTTPException],
                      Callable[[HTTPException], Response]]] = []


class ConfigurationError(InternalServerError):
    """Raised when a configuration parameter is missing or invalid."""


def handler(exception_type: Type[HTTPException]) -> Callable:
    """Generate a decorator to register a handler for an exception."""
    def deco(func: Callable) -> Callable:
        """Register a function as an exception handler."""
        _handlers.append((exception_type, func))
        return func
    return deco


def get_handlers() -> List[Tuple[Type[HTTPException], Callable]]:
    """
    Get a list of registered exception handlers.

    Returns
    -------
    list
        List of (:class:`.HTTPException`, callable) tuples.

    """
    return _handlers


def handle_exception(error: HTTPException) -> Response:
    """Render a generic error handler."""
    rendered = render_template("base/exception.html", error=error,
                               pagetitle=f"{error.code} {error.name }")
    response: Response = make_response(rendered, error.code)
    return response


# Generate handlers programmatically from the built-in Werkzeug HTTP
# exceptions.
for code, exception_type in default_exceptions.items():
    handler(exception_type)(handle_exception)


"""Tests for :mod:`arxiv.identifier`."""

from unittest import TestCase

from . import parse_arxiv_id


class TestParseArXivID(TestCase):
    """Tests for :func:`parse_arxiv_id`."""

    def test_old_style_with_archive(self):
        """Parse old-style arXiv IDs with archive names."""
        self.assertIsNotNone(parse_arxiv_id('math/9901123'))
        self.assertIsNotNone(parse_arxiv_id('hep-ex/9901123'))
        self.assertIsNotNone(parse_arxiv_id('gr-qc/9901123'))

    def test_old_style_with_category(self):
        """Parse old-style arXiv IDs with category names."""
        self.assertIsNotNone(parse_arxiv_id('math.NA/9901123'))
        self.assertIsNotNone(parse_arxiv_id('cs.DB/9901123'))

    def test_old_style_with_version(self):
        """Parse old-style arXiv IDs with version numbers."""
        self.assertIsNotNone(parse_arxiv_id('gr-qc/9901123v3'))

    def test_odd_mashup(self):
        """Parse identifer that combines old- and new-style."""
        self.assertIsNotNone(parse_arxiv_id('hep-ph/1203.12345v12'))

    def test_new_style(self):
        """Parse new-style identifiers."""
        self.assertIsNotNone(parse_arxiv_id('1202.1234'))
        self.assertIsNotNone(parse_arxiv_id('1203.12345'))

    def test_new_style_with_version(self):
        """Parse new-style with version numbers."""
        self.assertIsNotNone(parse_arxiv_id('1202.1234v1'))
        self.assertIsNotNone(parse_arxiv_id('1203.12345v1'))
        self.assertIsNotNone(parse_arxiv_id('1203.12345v12'))

    # slightly odd but seen in comments
